@@ -1,5 +1,7 @@
 package Parse::QTEDI;
 
+# Author: Dongxu Ma <dongxu@cpan.org>
+
 use 5.005;
 use strict;
 use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $parser $DEBUG);
@@ -9,9 +11,9 @@ require Exporter;
 @EXPORT_OK = qw($parser);
 
 use Parse::RecDescent ();
-use YAML ();
+use YAML::Syck ();
 
-$VERSION = '0.02_02';
+$VERSION = '0.19';
 $VERSION = eval $VERSION;  # see L<perlmodstyle>
 
 # Global flags 
@@ -44,8 +46,10 @@ __DATA__
 # CAUTION: the biggest assert here is we are working on a _VALID_ header
 begin          : <rulevar: local $stash = [] >
 begin          : 
-  loop(s) eof { print YAML::Dump($stash) } 
+  loop(s) eof { print YAML::Syck::Dump($stash) } 
 eof            : /^\Z/
+# make sure function_pointer is IN FRONT OF function
+# since function is compatible with function_pointer, unfortunately 
 primitive_loop : 
     qt_macro(s) 
   | kde_macro(s) 
@@ -56,6 +60,7 @@ primitive_loop :
   | extern(s)  
   | namespace(s) 
   | class(s) 
+  | function_pointer(s)
   | function(s) 
   | expression(s) 
 # inside a class each primitive code block has to consider 
@@ -69,7 +74,8 @@ primitive_loop_inside_class :
   | template 
   | extern 
   | namespace 
-  | class 
+  | class   
+  | function_pointer 
   | function 
   | expression 
 loop           : 
@@ -173,13 +179,14 @@ namespace:
   { foreach my $a (@{$item[4]}) { push @{$return->{body}}, @$a }    }
   { print STDERR "namespace: ", $item[2], "\n" if $::RD_DEBUG }
 class    : 
-  keyword_class class_name class_inheritance class_body variables ';'
+  keyword_class class_name class_inheritance class_body class_attribute variables ';'
   { 
     $return = { type => $item[1][0], name => $item[2] };
     $return->{property} = $item[1][1] if @{$item[1][1]};
     $return->{inheritance} = $item[3] if $item[3];
     $return->{body} = $item[4] if $item[4];
-    $return->{variable} = $item[5] if $item[5];
+    $return->{attribute} = $item[5] if $item[5];
+    $return->{variable} = $item[6] if $item[6];
     unless($item[4]) { 
         # no class body, possibly inside typedef or forward decl
         # split variables from enum_name
@@ -193,12 +200,17 @@ class    :
   { print STDERR $item[1][0], ": ", 
         join(" ", @item[2 .. $#item-3]), "\n" if $::RD_DEBUG }
 # a simple trap here 
-# to prevent template function parsed as normal one
+# to prevent template function from being parsed as normal one
 function : 
     keyword_template <commit> <reject>
   | class_accessibility <commit> <reject> 
   | function_header function_body
-    { $return = { type => 'function', %{$item[1]} }; } 
+    { 
+      $return = { type => 'function', %{$item[1]} }; 
+      if ($item[2]) {
+        push @{$return->{property}}, $item[2];
+      }
+    } 
     { print STDERR "function: ", $item[1]->{name}, "\n" if $::RD_DEBUG }
 # QT-specific macros
 qt_macro_1 : 
@@ -216,18 +228,20 @@ qt_macro_2 :
   'Q_DECLARE_MUTABLE_SEQUENTIAL_ITERATOR' | 
   'Q_DUMMY_COMPARISON_OPERATOR' 
 qt_macro_3 : 
-  'Q_PRIVATE_SLOT' | 'Q_PROPERTY'
+  'Q_PRIVATE_SLOT' | 'Q_PROPERTY' | 'Q_PRIVATE_PROPERTY' | 'Q_CLASSINFO' | 'Q_INTERFACES'
+qt_macro_10: 
+ 'Q_INVOKABLE' 
 qt_macro_99: 
   'Q_REQUIRED_RESULT' 
 qt_macro : 
-    qt_macro_1 
+    qt_macro_1 ( ';' | )
     { $return = { type => 'macro', subtype => 1, name => $item[1] } } 
     { print STDERR $item[1], "\n" if $::RD_DEBUG } 
-  | qt_macro_2 '(' next_end_bracket ')' 
+  | qt_macro_2 '(' next_end_bracket ')' ( ';' | )
     { $return = { type => 'macro', subtype => 2, name => $item[1], 
         value => $item[3] } } 
     { print STDERR join(" ", @item[1 .. 4]), "\n" if $::RD_DEBUG } 
-  | qt_macro_3 '(' balanced_bracket(s) ')' 
+  | qt_macro_3 '(' balanced_bracket(s) ')' ( ';' | )
     { $return = { type => 'macro', subtype => 3, name => $item[1], 
         values => join(" ", @{$item[3]}) } } 
     { print STDERR join(" ", $item[1 .. 2], @{$item[3]}, $item[4]), "\n" if $::RD_DEBUG } 
@@ -241,14 +255,14 @@ kde_macro_3  :
 kde_macro_99 : 
   'KDE_DEPRECATED' | 'KDE_EXPORT' 
 kde_macro : 
-    kde_macro_1 
+    kde_macro_1 ( ';' | )
     { $return = { type => 'macro', subtype => 1, name => $item[1] } } 
     { print STDERR $item[1], "\n" if $::RD_DEBUG } 
-  | kde_macro_2 '(' next_end_bracket ')' 
+  | kde_macro_2 '(' next_end_bracket ')' ( ';' | )
     { $return = { type => 'macro', subtype => 2, name => $item[1], 
         value => $item[3] } } 
     { print STDERR join(" ", @item[1 .. $#item]), "\n" if $::RD_DEBUG } 
-  | kde_macro_3 '(' balanced_bracket(s) ')' 
+  | kde_macro_3 '(' balanced_bracket(s) ')' ( ';' | )
     { $return = { type => 'macro', subtype => 3, name => $item[1], 
         values => join(" ", @{$item[3]}) } } 
     { print STDERR join(" ", $item[1 .. 2], @{$item[3]}, $item[4]), "\n" if $::RD_DEBUG } 
@@ -294,8 +308,10 @@ next_begin_or_end_bracket :
   /(?>[^\(\)]+)/sio   { ( $return = $item[1] ) =~ s/\n/ /go } 
 next_bracket_or_brace_or_semicolon : 
   /(?>[^\(\)\{\}\;]+)/sio { ( $return = $item[1] ) =~ s/\n/ /go } 
-next_bracket_or_square_bracket_or_brace_or_semicolon : 
-  /(?>[^\(\)\{\}\[\]\;]+)/sio { ( $return = $item[1] ) =~ s/\n/ /go } 
+next_bracket_or_brace_or_semicolon_or_equal : 
+  /(?>[^\(\)\{\}\;\=]+)/sio { ( $return = $item[1] ) =~ s/\n/ /go } 
+next_bracket_or_square_bracket_or_brace_or_semicolon_or_equal : 
+  /(?>[^\(\)\{\}\[\]\;\=]+)/sio { ( $return = $item[1] ) =~ s/\n/ /go } 
 next_semicolon : 
   /(?>[^\;]+)/sio     { ( $return = $item[1] ) =~ s/\n/ /go }
 next_begin_brace_or_colon_or_semicolon : 
@@ -326,21 +342,25 @@ balanced_angle_bracket :
 # array declaration should be handled carefully
 # FIXME: __attribute__()
 expression_next_token : 
-    next_bracket_or_square_bracket_or_brace_or_semicolon 
+    next_bracket_or_square_bracket_or_brace_or_semicolon_or_equal 
     { $return = $item[1] }
   | { $return = ''       } 
 array_dimention_next_token : 
     next_square_bracket { $return = $item[1] } 
   | { $return = ''       } 
 expression_body : 
-  expression_next_token array_dimention(s?) expression_value(?) 
-  { $return = join(" ", $item[1], @{$item[2]}) } 
+  expression_next_token 
+  { $item[1] =~ m/\boperator\W?$/o ? undef : 1 } 
+  array_dimention(s?) 
+  expression_value(?) 
+  { $return = join(" ", $item[1], @{$item[3]}, $item[4]) } 
 array_dimention : 
   '[' (   next_square_bracket { $return = $item[1] } 
         | { $return = '' } ) ']' 
   { $return = join(" ", @item[1 .. 3]) } 
 expression_value: 
   '=' ( '{' balanced_brace(s) '}' | next_semicolon ) 
+  { $return = '= '. $item[2] }
 # variable related
 variables : next_semicolon { $return = $item[1] } | { $return = '' } 
 
@@ -350,15 +370,18 @@ variables : next_semicolon { $return = $item[1] } | { $return = '' }
 function_header       : 
     (   keyword_comment | keyword_class | keyword_enum 
       | keyword_typedef ) <commit> <reject>
-  | function_header_block(s) 
+  | ( qt_macro_10 | ) function_header_block(s) 
     { 
       $return->{name} = ''; 
-      foreach my $i (@{$item[1]}) { 
+      my $seen_function_name = 0; 
+      foreach my $i (@{$item[2]}) { 
           if ($i->{_subtype} == 1) { 
               # attribute
-              $return->{name} .= $i->{_value}; 
+              # stripped currently
+              #$return->{name} .= " ". $i->{_value}; 
           } elsif ($i->{_subtype} == 2) { 
               # function name with params
+              $seen_function_name = 1;
               $return->{name} .= $i->{_name};
               $return->{parameter} = $i->{_value} if 
                   @{$i->{_value}};
@@ -366,56 +389,78 @@ function_header       :
               # other macros
               push @{$return->{property}}, $i->{_value} if 
                   $i->{_value};
-          } 
+          } elsif ($i->{_subtype} == 4) {
+              # const
+              if ($i->{_value}) {
+                  if ($seen_function_name) {
+                      push @{$return->{property}}, $i->{_value};
+                  }
+                  else {
+                      $return->{name} = $i->{_value}. ' '. $return->{name};
+                  }
+              }
+          }
       } 
     } 
 function_header_next_token : 
-  next_bracket_or_brace_or_semicolon 
+  next_bracket_or_brace_or_semicolon_or_equal 
     { $return = $item[1] } 
   | { $return = ''       } 
-# old backup 
-#function_header_block_old_unused : 
-#  function_header_next_token '(' function_header_loop(s?) ')' 
-#    { $return = join("", $item[1], $item[2], @{$item[3]}, $item[4]) }
-#  | function_macro_99(s?) 
-#    { $return = join("", @{$item[1]}) } 
 
 # function parameter process
 # parse parameters and simply consume __attribute__ 
 function_header_block : 
-  (   function_header_next_token 
+  (   /const\b/o
+      { 
+        #print STDERR "const\n";
+        $return = { _subtype => 4, _value => 'const' };
+      }
+    | function_header_next_token 
+      { $item[1] =~ m/\boperator\b/o ? 1 : undef } 
+      ( '(' ')' { $return = '()' } | next_begin_bracket { $return = $item[1] } )
+      '(' 
+      ( function_parameters { $return = $item[1]; } | { $return = []; } ) 
+      ')' 
+      { 
+        #print STDERR "operator\n";
+        $return = { _subtype => 2, _name => $item[1].$item[3], }; 
+        $return->{_value} = $item[5];
+      } 
+    | function_header_next_token 
+      { $item[1] =~ m/^\s*throw\s*$/o ? 1 : undef } 
+      '(' next_bracket_or_brace_or_semicolon ')'
+      { 
+        #print STDERR "throw\n";
+        $return = { _subtype => 3, _value => 'throw('. $item[4]. ')' }
+      }
+    | function_header_next_token 
       { $item[1] =~ m/\_\_attribute\_\_\s*$/o ? 1 : undef } 
       '(' function_header_loop(s?) ')' 
       { 
+        #print STDERR "__attribute__\n";
         $return = { _subtype => 1, 
                     _value   => join("", $item[1], $item[3], 
-                                    @{$item[4]}, $item[5]) }; 
+                                    @{$item[4]}, $item[5]) }
       } 
     | function_header_next_token 
       { $item[1] =~ m/^\:/o ? 1 : undef } 
       function_header_loop(s) 
-      { $return = { _subtype => 0 }; } 
-    | function_header_next_token 
-      { $item[1] =~ m/operator$/o ? 1 : undef } 
-      '(' ')' 
-      '(' 
-      ( function_parameters { $return = $item[1]; } | { $return = []; } ) 
-      ')' 
-      { 
-        $return = { _subtype => 2, _name => $item[1].'()', }; 
-        $return->{_value} = $item[6]; 
-      } 
+      { $return = { _subtype => 0 } } 
     | function_header_next_token 
       '(' 
       ( function_parameters { $return = $item[1]; } | { $return = []; } ) 
       ')' 
       { 
+        #print STDERR "name:$item[1]\n";
         $return = { _subtype => 2, _name => $item[1], };  
-        $return->{_value} = $item[3]; 
+        $return->{_value} = $item[3];
       } 
   ) { $return = $item[1]; } 
   | function_macro_99(s?) 
-    { $return = { _subtype => 3, _value => join("", @{$item[1]}) }; } 
+    { 
+      #print STDERR "property:", @{$item[1]}, "\n";
+      $return = { _subtype => 3, _value => join("", @{$item[1]}) }
+    } 
 # TODO: loop in more elegant way
 function_header_loop  : 
   (   function_header_next_token { $return = $item[1]; } 
@@ -460,6 +505,8 @@ function_parameter_declaration            :
         { $return = { subtype => 'fpointer', value => $item[1] }; } 
       | function_parameter_template_type 
         { $return = { subtype => 'template', value => $item[1] }; } 
+      | function_parameter_array_pointer 
+        { $return = { subtype => 'apointer', value => $item[1] }; }
       | { $return = { subtype => 'simple' }; } 
     ) 
     { 
@@ -470,11 +517,20 @@ function_parameter_declaration            :
           $return->{subtype} = 'template'; 
       } elsif ($item[2]->{subtype} eq 'fpointer') { 
           $return = $item[2]->{value};
-          $return->{name} = $item[1];
+          $return->{return}  = $item[1];
           $return->{subtype} = 'fpointer'; 
-      } 
+      } elsif ($item[2]->{subtype} eq 'apointer') {
+          $return = $item[2]->{value};
+          $return->{type} = $item[1]. $return->{type};
+          $return->{subtype} = 'apointer';
+      }
     } 
 
+function_parameter_array_pointer : 
+  '(' '&' function_parameter_declaration_next_token ')' 
+  '[' /(?>[^\]]+)/iso ']' 
+  { $return = { type => join("", @item[1,2,4..7]), name => $item[3] } }
+
 function_parameter_template_type_next_token : 
     /(?>[^\<\>]+)/iso { ( $return = $item[1] ) =~ s/\n/ /go } 
 
@@ -485,8 +541,9 @@ function_parameter_template_type            :
       | { $return = '';       } 
     ) 
     { 
-      $return = join(" ", $item[1], 
-          join("", @{$item[2]}), $item[3], $item[4]); 
+      $return = join(" ", 
+          join("", $item[1], join("", @{$item[2]}), $item[3]), 
+          $item[4]); 
     } 
 
 function_parameter_template_type_loop       : 
@@ -495,7 +552,7 @@ function_parameter_template_type_loop       :
       | { $return = '';       } 
     ) 
     (   '<' function_parameter_template_type_loop '>' 
-        { $return = join(" ", @item[1 .. 3]); } 
+        { $return = join("", @item[1 .. 3]); } 
       | { $return = '';                      } 
     ) 
     { $return = join("", @item[1 .. 2]); } 
@@ -505,8 +562,21 @@ function_parameter_function_pointer_next_token :
 
 function_parameter_function_pointer            : 
     '(' function_parameter_function_pointer_loop ')' 
+    { $item[2] ? 1 : undef }
     '(' function_parameter_loop ')' 
-    { $return = { name => $item[2], parameter => $item[5], }; }
+    function_parameter_function_pointer_const
+    { 
+      $return = { name => $item[2], parameter => $item[6], }; 
+      push @{ $return->{property} },  $item[8] if $item[8]; 
+    } 
+
+# 'const' could be either return/param type attribute
+# or function property
+function_parameter_function_pointer_const      : 
+    (   'const' { $return = 'const'; } 
+      | { $return = ''; } 
+    ) 
+    { $return = $item[1]; } 
 
 function_parameter_function_pointer_loop       : 
     (   function_parameter_function_pointer_next_token 
@@ -538,8 +608,12 @@ function_parameter_default_value            :
 function_parameter_default_value_loop_token_dispatch : 
     '(' ')' 
     { $return = '()'; } 
+  | '" "' { $return = $item[1]; }
+  | "' '" { $return = $item[1]; }
+  | "(' ')" { $return = $item[1]; }
+  | '(" ")' { $return = $item[1]; }
   | '(' function_parameter_default_value_loop2 ')' 
-    { $return = join(" ", @item[1 .. 3]); } 
+    { $return = join("", @item[1 .. 3]); } 
   | "'" /(?>[^\']*)/iso "'" { $return = join("", @item[1 .. 3]); } 
   | '"' /(?>[^\"]*)/iso '"' { $return = join("", @item[1 .. 3]); } 
 
@@ -551,7 +625,7 @@ function_parameter_default_value_loop2      :
     (   function_parameter_default_value_loop_token_dispatch 
         { $return = $item[1]; } 
       | ',' function_parameter_default_value_loop2 
-        { $return = join(" ", @item[1 .. 2]); } 
+        { $return = join("", @item[1 .. 2]); } 
       | { $return = '';                       } 
     ) 
     { $return = join("", @item[1 .. 2]); } 
@@ -569,7 +643,7 @@ function_parameter_default_value_loop       :
 
 function_body         : 
     ';' { $return = '' } 
-  | '=' '0' ';' { $return = '' }
+  | '=' '0' ';' { $return = 'pure virtual' }
   | '{' balanced_brace(s) '}' ( ';' | ) { $return = '' }
 balanced_brace_next_token : 
     next_begin_or_end_brace { $return = $item[1] }
@@ -586,12 +660,13 @@ enum_name          :
 enum_body          : 
     '{' '}' { $return = [] }
   | '{' enum_unit(s) '}'
-    { $return = $item[2] }
+    { $return = [ grep { ref $_ } @{$item[2]} ] }
     #{ print STDERR "enum_body: ", join(" ", @{$item[2]}), "\n" if $::RD_DEBUG } 
   | { $return = ''       }
 enum_unit          : 
-  next_dot_or_end_brace ( ',' | )
-  { $return = (split /=/, $item[1])[0] }
+    comment { $return = '' }
+  | next_dot_or_end_brace ( ',' | )
+    { $return = [ split /\s*=\s*/, $item[1] ] }
   #{ print STDERR "enum_unit: $return\n" if $::RD_DEBUG } 
 
 # template related
@@ -629,25 +704,29 @@ class_body          :
     { $return = $item[2] } 
   | { $return = ''       } 
 class_body_content  : 
-    class_accessibility 
-    { $return = $item[1] } 
-  | primitive_loop_inside_class
-    { $return = $item[1] } 
+    class_accessibility { $return = $item[1] } 
+  | noop(s) { $return = { type => 'noop' } } 
+  | primitive_loop_inside_class { $return = $item[1] } 
     #{ print STDERR "class_body_content: ", $return, "\n" if $::RD_DEBUG }
 #  | { $return = ''       } 
 #    #{ print STDERR "class_body_content: NULL\n" if $::RD_DEBUG } 
 class_accessibility_loop : 
-  class_accessibility_content | qt_accessibility_content | 
-  kde_accessibility_content 
+    ( class_accessibility_content { $return = $item[1] } | { $return = '' } ) qt_accessibility_content { $return = $item[1] ? join(' ', $item[1], $item[2]) : $item[2] }
+  | class_accessibility_content
+  | kde_accessibility_content 
 class_accessibility : 
-  class_accessibility_loop(s) ':' 
+  class_accessibility_loop ':' 
   { $return = { type => 'accessibility', value => $item[1] } } 
 qt_accessibility_content : 
-  'Q_SIGNALS' | 'Q_SLOTS' | 'signals' | 'slots' 
+  'Q_SIGNALS' | 'Q_SLOTS' | 'signals' | 'slots'
 kde_accessibility_content: 
   'k_dcop' 
 class_accessibility_content : 
   'public' | 'private' | 'protected' 
+class_attribute: 
+  /__attribute__\s*\(\((.+?)\)\)/io { $return = $1 } | { $return = '' }
+noop :
+  ';'
 
 #namespace related
 namespace_name : 
